"use strict";

var _interopRequireDefault = require("@babel/runtime-corejs3/helpers/interopRequireDefault");

var _regenerator = _interopRequireDefault(require("@babel/runtime-corejs3/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime-corejs3/helpers/asyncToGenerator"));

var _typeof2 = _interopRequireDefault(require("@babel/runtime-corejs3/helpers/typeof"));

var _forEach = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/for-each"));

var _keys = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/object/keys"));

var _isArray = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/array/is-array"));

var _slice = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/slice"));

var _promise = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/promise"));

var _indexOf = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/index-of"));

var _some = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/some"));

var _concat = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/concat"));

var _map = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/map"));

var _filter = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/filter"));

var _fs = _interopRequireDefault(require("fs"));

var _path = _interopRequireDefault(require("path"));

var _yaml = _interopRequireDefault(require("yaml"));

var DEFAULT_ARGV_START_POSITION = 2,
    RE_YAML = /^\.ya?ml$/i;
/**
 * Primary module
 **/

module.exports = function (settings) {
  var defaultOptions = {
    baseSettingsPath: '',
    commandLineSwitches: ['--config-file'],
    environmentSearchPaths: ['./', './config', './settings'],
    readCommandLineMap: {},
    readEnvironmentMap: {},
    strict: false
  },
      defaultTypesMap = {};
  /**
   * Creates an internal map used to understand the original value type of
   * a field within the base configuration file. Values of subsequent overrides
   * are coerced accordingly based on the map created.
   *
   * @param {object} source - an object
   * @param {string} parentPath - a namespaced field name used as a hash key
   * @returns {undefined}
   **/

  function buildDefaultTypesMap(source, parentPath) {
    var _context;

    (0, _forEach["default"])(_context = (0, _keys["default"])(source)).call(_context, function (key) {
      var keyPath = parentPath ? [parentPath, key].join('.') : key;

      if (source[key] === null) {
        return;
      } // array coercion


      if ((0, _isArray["default"])(source[key])) {
        defaultTypesMap[keyPath] = function (value) {
          if ((0, _isArray["default"])(value)) {
            return value;
          } // remove front and back brackets, then split on commas


          return (0, _slice["default"])(value).call(value, 1, -1).split(',');
        };

        return;
      } // date coercion


      if (Object.prototype.toString.call(source[key]) === '[object Date]') {
        defaultTypesMap[keyPath] = function (value) {
          return new Date(value);
        };

        return;
      } // object coercion


      if ((0, _typeof2["default"])(source[key]) === 'object' && /^\[object\sObject\]$/.test(source[key].toString()) && source[key] !== null) {
        return buildDefaultTypesMap(source[key], keyPath);
      }

      defaultTypesMap[keyPath] = function (value) {
        switch ((0, _typeof2["default"])(source[key])) {
          case 'boolean':
            return /^true|1$/i.test(value);

          case 'number':
            return Number(value);

          default:
            return value;
        }
      };

      return;
    });
  }
  /**
   * Returns a Promise that can be used to determine if a file exists.
   *
   * @param {string} filePath - a path to a file
   * @returns {Promise} - a native Promise that resolves true or false
   **/


  function checkIfFileExists(filePath) {
    return new _promise["default"](function (resolve) {
      return _fs["default"].exists(filePath, resolve);
    });
  }
  /**
   * Creates a new object and merges multiple objects together,
   * each subsequent object having priority
   *
   * @param {Array} objectList - accepts array of objects
   * @param {boolean} stripEmptyObjects - when true, removes empty objects from cloned result
   * @returns {Object} a newly cloned object with the merged results
   **/


  function cloneAndMerge(objectList, stripEmptyObjects) {
    if (!(0, _isArray["default"])(objectList)) {
      return objectList;
    }

    var checkStrict = settings.options && settings.options.baseSettingsPath && settings.options.strict,
        cloned = {},
        cloner = function cloner(source, destination, prefix, ensureKey) {
      var _context2;

      (0, _forEach["default"])(_context2 = (0, _keys["default"])(source)).call(_context2, function (key) {
        var objectKeyPath = prefix ? [prefix, key].join('.') : key;

        if ((0, _isArray["default"])(source[key])) {
          destination[key] = source[key];
        } else if ((0, _typeof2["default"])(source[key]) === 'object' && source[key] !== null) {
          destination[key] = cloner(source[key], destination[key] || {}, objectKeyPath, ensureKey); // remove empty objects when applicable...

          if (!(0, _keys["default"])(destination[key]).length && stripEmptyObjects) {
            delete destination[key];
          }
        } else if (!ensureKey || isKeyInBase(objectKeyPath)) {
          destination[key] = source[key];
        }
      });
      return destination;
    };

    (0, _forEach["default"])(objectList).call(objectList, function (item, i) {
      if ((0, _typeof2["default"])(item) === 'object') {
        cloned = cloner(item, cloned, '', i !== 0 && checkStrict);
      }
    });
    return cloned;
  }
  /**
   * Creates a new object based on an expression representing a single field
   * in an object and applies specified value
   *
   * @param {string} expression - string Javascript expression
   * @param {Array|Boolean|Number|Object|string} value - value to assign to the end result of the expression
   * @returns {Object} - the object created
   **/


  function createObjectWithValue(expression, value) {
    var atEnd = function atEnd(list, key) {
      return (0, _indexOf["default"])(list).call(list, key) === list.length - 1;
    },
        current,
        keys = [],
        // matches ], '], "], ][, '][', "][", [, [' and ["
    regexHash = /([\'\"]?\]\[[\'\"]?)|(\[[\'\"]?)|([\'\"]?\])/g,
        stack = {},
        transform;

    if (expression) {
      var _context3;

      // breaks the expression into an array which represents its stack
      (0, _forEach["default"])(_context3 = expression.split(regexHash)).call(_context3, function (key) {
        if (key && !key.match(regexHash)) {
          keys.push(key);
        }
      }); // in the event hash notation was not used, but periods were...

      if (keys.length === 1) {
        keys = keys[0].split('.');
      } // define value transform function


      transform = defaultTypesMap[keys.join('.')] || function (input) {
        return input;
      }; // create an object graph representing string based script notation


      (0, _forEach["default"])(keys).call(keys, function (key) {
        if (!current) {
          current = stack;
        }

        current[key] = atEnd(keys, key) ? transform(value) : {};
        current = current[key];
      });
    }

    return stack;
  }
  /**
   * Loads JSON configuration data from a text file
   *
   * @param {string} filePath - path to JSON configuration file
   * @returns {Promise} - a native Promise that resolves the file content
   **/


  function getConfigFileContents(filePath) {
    return new _promise["default"](function (resolve, reject) {
      var chunks = [],
          data,
          fileRead = _fs["default"].createReadStream(filePath, {
        encoding: 'utf8'
      });

      fileRead.on('data', function (chunk) {
        return chunks.push(chunk);
      });
      fileRead.on('end', function () {
        var content = chunks.join(''); // attempt to load and parse if it is a YAML file

        if (RE_YAML.test(_path["default"].extname(filePath))) {
          try {
            data = _yaml["default"].parse(content, {
              merge: true
            });
          } catch (ex) {
            return reject(new Error("settings-lib: unable to parse YAML settings: ".concat(ex.message)));
          }

          return resolve(data);
        } // otherwise, attempt to parse as JSON


        try {
          data = JSON.parse(content);
        } catch (ex) {
          return reject(new Error("settings-lib: unable to parse JSON settings: ".concat(ex.message)));
        }

        return resolve(data);
      });
      fileRead.on('error', function (err) {
        return reject(new Error("settings-lib: unable to read settings file: ".concat(err.message)));
      });
    });
  }

  function isKeyInBase(key) {
    return typeof defaultTypesMap[key] !== 'undefined';
  }
  /**
   * Loads base configuration for settings
   *
   * @returns {Promise} - a native Promise
   **/


  function loadBaseConfig() {
    return new _promise["default"](function (resolve, reject) {
      if (!settings.options.baseSettingsPath) {
        settings.baseConfig = {};
        return resolve();
      }

      return getConfigFileContents(settings.options.baseSettingsPath).then(function (contents) {
        buildDefaultTypesMap(contents);
        settings.baseConfig = contents;
        return resolve();
      })["catch"](reject);
    });
  }
  /**
   * Loads command line configuration file
   *
   * @returns {Promise} - a native Promise
   **/


  function loadCommandLineConfig() {
    return new _promise["default"](function (resolve, reject) {
      var _context4;

      var args = (0, _slice["default"])(_context4 = process.argv).call(_context4, DEFAULT_ARGV_START_POSITION),
          // removes node command
      commandLineConfigPath; // no sense in doing this routine if no command line switches are configured

      if (!(0, _isArray["default"])(settings.options.commandLineSwitches) || settings.options.commandLineSwitches.length < 1) {
        return resolve();
      } // iterate args from command line


      (0, _some["default"])(args).call(args, function (arg, i) {
        var _context5;

        if (!arg) {
          return false;
        }

        if ((0, _indexOf["default"])(_context5 = settings.options.commandLineSwitches).call(_context5, arg) >= 0 && args[i + 1]) {
          commandLineConfigPath = args[i + 1];
          return true;
        }

        return false;
      }); // if no path was identified, exit out

      if (!commandLineConfigPath) {
        return resolve();
      } // if a path was identified, load it


      return getConfigFileContents(commandLineConfigPath).then(function (contents) {
        settings.commandLineConfig = contents;
        return resolve();
      })["catch"](reject);
    });
  }
  /**
   * Applies command line switch overrides
   *
   * @returns {Promise} - a native Promise
   **/


  function loadCommandLineSwitches() {
    return new _promise["default"](function (resolve) {
      var _context6, _context7;

      var args = (0, _slice["default"])(_context6 = process.argv).call(_context6, DEFAULT_ARGV_START_POSITION),
          // removes node command
      commandLineMapKeys = (0, _keys["default"])(settings.options.readCommandLineMap),
          commandLineSettings = []; // return if there are no configured command line mappings to look for

      if (!commandLineMapKeys.length) {
        return resolve();
      } // iterate each arg and associate to any configured command line overrides


      (0, _forEach["default"])(args).call(args, function (arg, i) {
        if (!arg) {
          return;
        }

        if ((0, _indexOf["default"])(commandLineMapKeys).call(commandLineMapKeys, arg) >= 0 && args[i + 1]) {
          // collect command line override value
          commandLineSettings.push(createObjectWithValue(settings.options.readCommandLineMap[arg], args[i + 1]));
        }
      });
      settings.commandLineConfig = cloneAndMerge((0, _concat["default"])(_context7 = [settings.commandLineConfig || {}]).call(_context7, commandLineSettings));
      return resolve();
    });
  }
  /**
   * Loads environment configuration file
   *
   * @returns {Promise} - a native Promise
   **/


  function loadEnvironmentConfig() {
    return new _promise["default"](function (resolve, reject) {
      var _context8, _context9, _context10;

      var searchEnvironment = process.env.NODE_ENV && settings.options.environmentSearchPaths && (0, _isArray["default"])(settings.options.environmentSearchPaths);

      if (!searchEnvironment) {
        return resolve();
      }

      return _promise["default"].all((0, _concat["default"])(_context8 = (0, _map["default"])(_context9 = settings.options.environmentSearchPaths // check for .json files...
      ).call(_context9, function (searchPath) {
        var environmentOverridePath = _path["default"].resolve([_path["default"].join(searchPath, process.env.NODE_ENV), '.json'].join(''));

        return new _promise["default"](function (resolve, reject) {
          return checkIfFileExists(environmentOverridePath).then(function (exists) {
            return resolve({
              exists: exists,
              path: environmentOverridePath
            });
          })["catch"](reject);
        });
      })).call(_context8, (0, _map["default"])(_context10 = settings.options.environmentSearchPaths // check for .yml files...
      ).call(_context10, function (searchPath) {
        var environmentOverridePath = _path["default"].resolve([_path["default"].join(searchPath, process.env.NODE_ENV), '.yml'].join(''));

        return new _promise["default"](function (resolve, reject) {
          return checkIfFileExists(environmentOverridePath).then(function (exists) {
            return resolve({
              exists: exists,
              path: environmentOverridePath
            });
          })["catch"](reject);
        });
      }))).then(function (resultsList) {
        var validPaths = (0, _filter["default"])(resultsList).call(resultsList, function (result) {
          return result.exists;
        });

        if (!validPaths.length) {
          return resolve();
        }

        return getConfigFileContents(validPaths[0].path).then(function (contents) {
          settings.environmentConfig = contents;
          return resolve();
        })["catch"](reject);
      })["catch"](reject);
    });
  }
  /**
   * Loads environment variable overrides
   *
   * @returns {Promise} - a native Promise
   **/


  function loadEnvironmentVariables() {
    return new _promise["default"](function (resolve) {
      var _context11;

      var environmentKeys,
          environmentMapKeys = (0, _keys["default"])(settings.options.readEnvironmentMap),
          environmentSettings = []; // return if there are no configured command line mappings to look for

      if (!environmentMapKeys.length) {
        return resolve();
      }

      environmentKeys = (0, _keys["default"])(process.env);
      (0, _forEach["default"])(environmentKeys).call(environmentKeys, function (key) {
        if ((0, _indexOf["default"])(environmentMapKeys).call(environmentMapKeys, key) >= 0) {
          environmentSettings.push(createObjectWithValue(settings.options.readEnvironmentMap[key], process.env[key]));
        }
      });
      settings.environmentConfig = cloneAndMerge((0, _concat["default"])(_context11 = [settings.environmentConfig || {}]).call(_context11, environmentSettings));
      return resolve();
    });
  }
  /**
   * Initializes settings, refreshes all config and begins load
   * from base file (if specified), environment and then command line
   *
   * @param {Object} options - an optional set of instructions for initialization
   * @param {function} callback - an optional callback function
   * @returns {undefined|Promise} - if callback is not supplied, a native Promise is returned
   **/


  settings.initialize = /*#__PURE__*/function () {
    var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(options, callback) {
      var exec;
      return _regenerator["default"].wrap(function _callee2$(_context13) {
        while (1) {
          switch (_context13.prev = _context13.next) {
            case 0:
              if (typeof options === 'function') {
                callback = options;
                options = {};
              } // ensure options are empty if null or not set


              options = options || {}; // handle backwards compatability for previous versions of library

              if (options.baseConfigPath) {
                options.baseSettingsPath = options.baseConfigPath;
                delete options.baseConfigPath;
              }

              exec = /*#__PURE__*/function () {
                var _ref2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
                  return _regenerator["default"].wrap(function _callee$(_context12) {
                    while (1) {
                      switch (_context12.prev = _context12.next) {
                        case 0:
                          // clean up and prepare internal references
                          delete settings.baseConfig;
                          delete settings.environmentConfig;
                          delete settings.commandLineConfig; // get input options sorted...

                          settings.options = cloneAndMerge([defaultOptions, options]); // start loading up and applying all layers of configuration

                          _context12.next = 6;
                          return loadBaseConfig();

                        case 6:
                          _context12.next = 8;
                          return loadEnvironmentConfig();

                        case 8:
                          _context12.next = 10;
                          return loadEnvironmentVariables();

                        case 10:
                          _context12.next = 12;
                          return loadCommandLineConfig();

                        case 12:
                          _context12.next = 14;
                          return loadCommandLineSwitches();

                        case 14:
                          // prepare the settings
                          settings.config = cloneAndMerge([settings.baseConfig || {}, settings.environmentConfig || {}, settings.commandLineConfig || {}], true);
                          return _context12.abrupt("return", settings.config);

                        case 16:
                        case "end":
                          return _context12.stop();
                      }
                    }
                  }, _callee);
                }));

                return function exec() {
                  return _ref2.apply(this, arguments);
                };
              }(); // return the async function / Promise


              if (callback) {
                _context13.next = 8;
                break;
              }

              _context13.next = 7;
              return exec();

            case 7:
              return _context13.abrupt("return", _context13.sent);

            case 8:
              return _context13.abrupt("return", exec().then(function (result) {
                return callback(null, result);
              })["catch"](function (ex) {
                return callback(ex);
              }));

            case 9:
            case "end":
              return _context13.stop();
          }
        }
      }, _callee2);
    }));

    return function (_x, _x2) {
      return _ref.apply(this, arguments);
    };
  }();

  return settings;
}({});
//# sourceMappingURL=settings.js.map
